/*             This file is part of the New World OS project
--                 Copyright (C) 2006-2007  QRW Software
--           J. Scott Edwards - j.scott.edwards.nwos@gmail.com 
--                      http://www.qrwsoftware.com
--                      http://nwos.sourceforge.com
--
-- NWOS is free software;  you can redistribute it and/or modify it under the
-- terms of the GNU General Public License  as published by the Free Software
-- Foundation; either version 2, or (at your option) any later version.  This
-- software is distributed with the hope that it will be useful,  but WITHOUT
-- ANY WARRANTY;  without  even the  implied warranty  of MERCHANTABILITY  or
-- FITNESS FOR A PARTICULAR PURPOSE.   See the GNU General Public License for
-- more  details.  You should have received a copy  of the GNU General Public
-- License along with this package;  see the file LICENSE.  If not, write to:
--
--      Free Software Foundation
--      51 Franklin Street, Fifth Floor
--      Boston, MA 02110-1301
--      USA
--      http://www.fsf.org/licenses
--
-- $Log: file.c,v $
-- Revision 1.66  2007/06/18 13:16:17  jsedwards
-- Fix so create_file_without_storing correctly adds file path separator.
--
-- Revision 1.65  2007/06/13 21:19:47  jsedwards
-- Change find_matching_path_and_file function to accept a directory path that
-- is null and just ignore it.
--
-- Revision 1.64  2007/06/13 21:12:56  jsedwards
-- Change to correctly append a '/' to the directory path if it is missing.
--
-- Revision 1.63  2007/06/13 20:49:36  jsedwards
-- Added directory parameter to find_matching_path_and_file_association.
--
-- Revision 1.62  2007/06/13 20:16:50  jsedwards
-- Added a directory parameter to create_file_without_storing_data.
--
-- Revision 1.61  2007/06/05 23:49:05  jsedwards
-- Modify to allow option to ignore time stamp on files.
--
-- Revision 1.60  2007/04/07 02:17:46  jsedwards
-- Added code to print if matching path and file objects were printed.
--
-- Revision 1.59  2007/03/18 19:54:20  jsedwards
-- Added file_has_backup routine.
--
-- Revision 1.58  2007/03/18 18:17:01  jsedwards
-- Moved freeing reference list outside of if so it was at the same level as
-- the malloc.
--
-- Revision 1.57  2007/02/28 14:04:44  jsedwards
-- Added "file_is_stored" function.
--
-- Revision 1.56  2007/01/27 17:10:36  jsedwards
-- Added file_is_identical function that compares the actual data stored in
-- the system with the file, not the checksums.
--
-- Revision 1.55  2007/01/27 15:46:08  jsedwards
-- Change name of check_file_md5sum to check_file_checksums and correct bytes
-- printed for SHA1 sum.
--
-- Revision 1.54  2007/01/27 15:38:34  jsedwards
-- Change to check SHA1 sum as well as MD5 when restoring file.
--
-- Revision 1.53  2007/01/27 15:24:22  jsedwards
-- Change to check SHA1 sum as well as MD5 when verifying file.
--
-- Revision 1.52  2007/01/27 14:55:45  jsedwards
-- Added code to handle the case where a duplicate file with a different name
-- already exists in the system.
--
-- Revision 1.51  2007/01/27 14:26:36  jsedwards
-- Added code to set reference to 'void' in all of the find routines in the
-- case that they don't find what they are looking for.
--
-- Revision 1.50  2007/01/27 01:56:13  jsedwards
-- Changed two print statements in find_matching_file_from_size_md5_sha1 to
-- only print if 'vebose' global variable is set.
--
-- Revision 1.49  2007/01/20 13:42:53  jsedwards
-- Change to use the SHA1 sum to search for the file instead of the MD5
-- (less likely to be two files with the same checksum?).  Then just use
-- the MD5 to verify.
--
-- Revision 1.48  2007/01/19 15:06:11  jsedwards
-- Fix bug in creating sha1 checksum for files without storing data.
--
-- Revision 1.47  2007/01/19 13:14:10  jsedwards
-- Fix bug in compare time stamp.
--
-- Revision 1.46  2007/01/17 13:11:36  jsedwards
-- Move push and pop encryption level so that if it finds an existing file
-- (and returns an error) it doesn't leave it pushed.
--
-- Revision 1.45  2007/01/14 22:20:00  jsedwards
-- Modifications to do SHA1 checksum in addition to MD5 checksum.
--
-- Revision 1.44  2007/01/14 17:10:39  jsedwards
-- Changed file_path_to_file to file_path_to_path_and_file_association and
-- changed restore_file to take a path_and_file_assocation object as input
-- instead of a file object.
--
-- Revision 1.43  2007/01/14 13:35:51  jsedwards
-- Change nwos_number_of_files_for_path to take a path object instead of a
-- character path so it doesn't have to search for it (redundant).
--
-- Revision 1.42  2007/01/14 13:23:33  jsedwards
-- Added number_of_file_for_path and find_path_and_file_association_by_index
-- functions.
--
-- Revision 1.41  2007/01/14 13:19:44  jsedwards
-- Changed find_path_and_file_association so mod_time cannot be null (mod_time
--  has to match) because otherwise it would only return the first one found
-- (no way to find subsequent matches).
--
-- Revision 1.40  2007/01/14 04:03:10  jsedwards
-- Move modification time from 'file' object to 'path_and_file_association'
-- object.
--
-- Revision 1.39  2007/01/13 20:14:10  jsedwards
-- Changed so that file id isn't generated unless file is going to be created.
-- Also changed so 'seq' variable in 'read_file' keeps incrementing instead of
-- being mod'ed itself, and it is moded when used to index into sequence table.
--
-- Revision 1.38  2007/01/12 13:58:27  jsedwards
-- Add code to print read_file messages to the log too.
--
-- Revision 1.37  2007/01/10 14:55:16  jsedwards
-- Separated out routine to read file from create_file.
--
-- Revision 1.36  2007/01/07 14:16:47  jsedwards
-- Changed file_path_to_file to deal with new path_and_file_association object,
-- however it can currently only deal with a path to one file relationship.
--
-- Revision 1.35  2007/01/07 03:19:39  jsedwards
-- Fixed create_file routine to work with new file association object.
--
-- Revision 1.34  2007/01/06 20:13:32  jsedwards
-- Changed to find_private_name instead of find_name.
--
-- Revision 1.33  2007/01/05 13:47:46  jsedwards
-- Fix uninitialized variables in find_matching_file_from_size_time_md5.
--
-- Revision 1.32  2007/01/05 13:37:38  jsedwards
-- Add find_matching_path_and_file_association function.
--
-- Revision 1.31  2007/01/04 19:17:09  jsedwards
-- Moved path and file association routines lower in file so they can use
-- routines that are declared farther down.
--
-- Revision 1.30  2007/01/04 19:08:26  jsedwards
-- Extracted routine to find a file from the size, time, and md5sum from the
-- create_file_without_storing_data function.
--
-- Revision 1.29  2007/01/04 18:18:32  jsedwards
-- Extracted loop from create_file_without_storing_data that checksums file
-- into a separate function.
--
-- Revision 1.28  2007/01/04 17:32:26  jsedwards
-- In find routines change to return false if private class doesn't exist and
-- in create routines change to find or create the private class.
--
-- Revision 1.27  2007/01/04 04:05:59  jsedwards
-- Fixed mispelled function names.
--
-- Revision 1.26  2007/01/03 15:44:17  jsedwards
-- Fixed path and file association stuff to compile (UNTESTED!)
--
-- Revision 1.25  2006/12/24 15:05:12  jsedwards
-- Added code to clear md5sum, disc_copy, and location objects to zero when
-- they are created.  This was somehow left out earlier and objects of these
-- types had garbage in the unused fields.
--
-- Revision 1.24  2006/12/21 13:14:28  jsedwards
-- Hack to make it compile with new split public and private classes,
-- NON-FUNCTIONAL!
--
-- Revision 1.23  2006/12/19 14:52:16  jsedwards
-- Kludged version to compile with new 'file path' structure without 'file'
-- feature.  DOES NOT function correctly!
--
-- Revision 1.22  2006/12/19 14:42:21  jsedwards
-- Kludged version to compile with new File_Path structure that doesn't have
-- the 'file' variable.  This version is BROKEN!
--
-- Revision 1.21  2006/12/07 14:08:46  jsedwards
-- Moved setup routine and associated functions to attic/big_bang.c.
--
-- Revision 1.20  2006/12/02 03:41:49  jsedwards
-- Changed create_file_path to return created_new when it added the file to
-- an existing path name.
--
-- Revision 1.19  2006/12/01 14:31:30  jsedwards
-- Changed to use new malloc_reference_list and free_reference_list functions
-- instead of inlining the code.
--
-- Revision 1.18  2006/12/01 05:19:14  jsedwards
-- Added code that updates the data checksum and re-writes the object to disk
-- when changing from a single file to a file list.
--
-- Revision 1.17  2006/11/30 14:59:44  jsedwards
-- Change to allow 'file' in a file path object to point to a reference list
-- of files, if there is more than one file with that name.
--
-- Revision 1.16  2006/11/29 18:49:49  jsedwards
-- Change so both path and file object reference are passed to restore_file.
-- Also changed file error handling to exit on errors (this is probably bad).
--
-- Revision 1.15  2006/11/28 13:13:29  jsedwards
-- Changed to use BSD name for modification time in stat struct and added a
-- define to map it to the Linux name when compiling on Linux.
--
-- Revision 1.14  2006/11/27 13:48:08  jsedwards
-- Changed to use variable sized counts for disc lists (so number of files can
-- be larger than 127).
--
-- Revision 1.13  2006/11/19 16:38:18  jsedwards
-- Added find_matching_disc_list function.
--
-- Revision 1.12  2006/11/19 15:28:20  jsedwards
-- Made get_disc_list_object_size a global function.
--
-- Revision 1.11  2006/11/19 14:48:47  jsedwards
-- Change check_file_md5sum function to return one of three possible results
-- (file not found, md5sum match, or md5sum mismatch) instead of just a
-- boolean saying file found or not found.
--
-- Revision 1.10  2006/11/18 15:09:09  jsedwards
-- Added "max_size" parameter to read_variable_sized_object_from_disk because
-- objects are no longer limited to one file block.
--
-- Revision 1.9  2006/11/18 14:33:03  jsedwards
-- Change size of kludge buffer for reading and writing disc lists to max size
-- of disc list instead of file block size, because now disc list objects can
-- be larger than one file block.
--
-- Revision 1.8  2006/11/07 14:07:36  jsedwards
-- Completely rearranged create_file_without_storing_data function to deal
-- with files that have duplicate md5sums (and are possibly exact duplicates).
-- Fixed bug where all of disc id was not being compared.
--
-- Revision 1.7  2006/11/06 13:51:24  jsedwards
-- Change so create_file_without_storing_data returns the reference for the
-- file_path instead of the file.  Also fixed bug in class name.
--
-- Revision 1.6  2006/11/05 21:29:51  jsedwards
-- Add functions for finding and creating disc_copy and storage_location
-- objects.
--
-- Revision 1.5  2006/11/04 18:59:21  jsedwards
-- Added routines to find or create a Disc_List.
--
-- Revision 1.4  2006/11/02 11:45:59  jsedwards
-- Change maximum file size from 4GB - 1 to 2GB - 1 because the old compiler
-- isn't happy about the 32 bit constant.  It needs to be addressed sometime
-- in the future anyway.
--
-- Revision 1.3  2006/10/29 13:28:40  jsedwards
-- Added routines to just record a files md5sum without storing the file and
-- to check the md5sum of a file.
--
-- Revision 1.2  2006/10/26 01:51:27  jsedwards
-- Merged alpha_05_branch back into main trunk.
--
-- Revision 1.1.2.24  2006/10/25 12:22:28  jsedwards
-- Changed C_struct_class_definition to C_struct_Class_Definition so the case
-- is consistent with all the other C_struct objects.
--
-- Revision 1.1.2.23  2006/10/22 13:11:55  jsedwards
-- Changed to pass approximate number of blocks needed for a file to
-- set_sequential_blocks instead of spacing.  It now computes the
-- spacing from the blocks needed because the blocks on disk is variable.
--
-- Revision 1.1.2.22  2006/10/18 13:10:32  jsedwards
-- Changed printf formats for uint32 that is an integer instead of long.
--
-- Revision 1.1.2.21  2006/10/07 22:25:21  jsedwards
-- Changed so that sequential blocks are handled within objectify when the
-- 'set_sequential_blocks' function is called first.
--
-- Revision 1.1.2.20  2006/10/06 04:42:49  jsedwards
-- Changed to use new id generator that generates ids that scan across the
-- disk drive instead of totally random so that files don't cause so much
-- repositioning of the heads and hopefully will be faster.
--
-- Revision 1.1.2.19  2006/09/18 11:38:12  jsedwards
-- Fix so block sequence number wraps correctly.
--
-- Revision 1.1.2.18  2006/09/18 01:34:23  jsedwards
-- Changed for 256 byte blocks and use a "reference_list" to point to them.
--
-- Revision 1.1.2.17  2006/09/01 13:27:20  jsedwards
-- Changed "nwos_object_size" to "nwos_reference_list_size" and added the
-- object reference to "nwos_fill_in_common_header" so it can put the "id"
-- in the header now.
--
-- Revision 1.1.2.16  2006/08/26 15:30:11  jsedwards
-- Added routines to find and create MD5 objects and modified file routines
-- to use them instead of having the md5sum embedded in the file object.
-- Also removed the "tolower" for path names, it was a mistake.
--
-- Revision 1.1.2.15  2006/08/25 01:37:47  jsedwards
-- Added code to test md5sum on restoring the file.
--
-- Revision 1.1.2.14  2006/08/24 13:01:18  jsedwards
-- Added code to compute the MD5 checksum while reading the file.
--
-- Revision 1.1.2.13  2006/08/24 12:58:08  jsedwards
-- Changed the <utime.h> include to <sys/time.h> because although the man page
-- says "utimes()" is defined in <utime.h>, gcc still gives a warning.  I only
-- see "utimes()" in <sys/time.h>.
--
-- Revision 1.1.2.12  2006/08/24 11:57:21  jsedwards
-- Added code to restore the modification time of the file when it is restored.
--
-- Revision 1.1.2.11  2006/08/23 13:23:09  jsedwards
-- Added code to save the files time stamp.
--
-- Revision 1.1.2.10  2006/08/23 12:41:56  jsedwards
-- Changed to use fread instead of fgetc for every byte.
--
-- Revision 1.1.2.9  2006/08/20 15:35:27  jsedwards
-- Removed many of the debugging printf statements.
--
-- Revision 1.1.2.8  2006/08/20 15:13:43  jsedwards
-- Fix MAX_BLOCKS define.
--
-- Revision 1.1.2.7  2006/08/20 14:54:01  jsedwards
-- Fixed bug in num_blocks calculation and reading small (< 28 bytes) files.
--
-- Revision 1.1.2.6  2006/08/20 02:17:53  jsedwards
-- Fixed several bugs in the create_file function and added the restore_file
-- function.
--
-- Revision 1.1.2.5  2006/08/19 20:02:55  jsedwards
-- Fix comment about create_file routine.
--
-- Revision 1.1.2.4  2006/08/19 18:58:18  jsedwards
-- Added code to store a file in objects.
--
-- Revision 1.1.2.3  2006/08/19 18:51:59  jsedwards
-- Changed so checksum uses new "file" element instead of "count" for it's base.
--
-- Revision 1.1.2.2  2006/08/19 14:34:29  jsedwards
-- Added creation of MD5sum class (moved here from big_bang.c).
--
-- Revision 1.1.2.1  2006/08/18 13:04:03  jsedwards
-- Initial version.
--
*/

#define _LARGEFILE64_SOURCE

#include <assert.h>
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <unistd.h>

#include "crc32.h"
#include "md5/md5.h"
#include "objectify.h"
#include "objectify_private.h"
#include "sha1/sha1.h"
#include "time_stamp.h"

/* Map linux file modification time name in stat struct */
#ifdef linux
#define st_mtimespec st_mtim
#endif

#define DATA_STORAGE_SIZE (FILE_BLOCK_SIZE - 12)   /* 4 bytes each for flags, id, checksum */

typedef struct file_block {
    uint8 flags[4];
    ObjRef id;
    uint8 checksum[4];
    uint8 storage[DATA_STORAGE_SIZE];
} DataObject;



static size_t get_path_object_size(void* file_path_obj)
{
    assert(((C_struct_File_Path*)file_path_obj)->count > 0);

    return sizeof(C_struct_File_Path) + ((C_struct_File_Path*)file_path_obj)->count;
}



/*-------------------------------------------------------------------------------------------------------------------*/
/* File path (name) object stuff */
/*-------------------------------------------------------------------------------------------------------------------*/

bool nwos_find_file_path(char* path, ObjRef* ref)
{
    C_struct_Class_Definition class_def_obj;
    uint8 kludge[FILE_BLOCK_SIZE];
    C_struct_File_Path* ptr_path_obj = (C_struct_File_Path*)kludge;
    ObjRef path_class_ref;
    ObjRef object_class;
    ReferenceList* ref_list;
    int num_refs;
    size_t length;
    int i;

    length = strlen(path);

    assert(length <= 255);

    void_reference(ref);   /* in case we don't find it */


    /* for now return false if file path private class doesn't exist yet */

    if (!nwos_find_private_class_definition("FILE PATH", &path_class_ref)) return false;


    nwos_read_class_definition(&path_class_ref, &class_def_obj);

    ref_list = nwos_malloc_reference_list(&class_def_obj.header.object.references);

    num_refs = ref_list->common_header.num_refs;

    /* printf("num_refs: %d\n", num_refs); */

    for (i = 0; i < num_refs; i++)
    {
	nwos_get_object_class(&ref_list->references[i], &object_class);

	if (is_same_object(&object_class, &path_class_ref))
	{
	    nwos_read_variable_sized_object_from_disk(&ref_list->references[i], kludge, sizeof(kludge), &get_path_object_size);

	    /* remember ptr_path_obj points to the kludge buffer */

	    if (ptr_path_obj->count == length && strncasecmp((char*)ptr_path_obj->storage, path, length) == 0)   /* found a match */
	    {
		memcpy(ref, &ref_list->references[i], sizeof(ObjRef));
		break;
	    }
	}
    }

    nwos_free_reference_list(ref_list);
    ref_list = NULL;

    return (i != num_refs);   /* return true if we found it */
}



/* returns -1 if file path isn't found, 0 if path was found but no associated files, otherwise number of associated files */
int nwos_number_of_files_for_path(ObjRef* path_ref)
{
    uint8 kludge[FILE_BLOCK_SIZE];
    C_struct_File_Path* ptr_path_obj = (C_struct_File_Path*)kludge;
    ObjRef assoc_class_ref;
    ObjRef object_class;
    ReferenceList* ref_list;
    int num_refs;
    int i;
    int result = 0;


    if (nwos_find_private_class_definition("PATH AND FILE ASSOCIATION", &assoc_class_ref))
    {
	nwos_read_variable_sized_object_from_disk(path_ref, kludge, sizeof(kludge), &get_path_object_size);

	/* remember ptr_path_obj points to the kludge buffer */

	ref_list = nwos_malloc_reference_list(&ptr_path_obj->header.object.references);

	num_refs = ref_list->common_header.num_refs;

	for (i = 0; i < num_refs; i++)
	{
	    nwos_get_object_class(&ref_list->references[i], &object_class);

	    if (is_same_object(&object_class, &assoc_class_ref)) result++;
	}

	nwos_free_reference_list(ref_list);
	ref_list = NULL;
    }

    return result;   /* return number of path_and_file_associations we found that refer to this path */
}



/* Find existing path or create new */

ObjCreateResult nwos_create_file_path(char* path, ObjRef* ref)
{
    uint8 kludge[FILE_BLOCK_SIZE];
    C_struct_File_Path* ptr_path_obj = (C_struct_File_Path*)kludge;
    size_t length;
    ObjRef path_class_ref;
    ObjCreateResult result = FOUND_EXISTING;
    int i;

    length = strlen(path);

    assert(length <= 255);
    
    /* first find out if we already have this path */

    if (!nwos_find_file_path(path, ref))   /* didn't find it */
    {
	memset(kludge, 0, sizeof(kludge));  /* zero it out */

	/* remember ptr_path_obj points to the kludge buffer */

	nwos_find_or_create_private_class_definition("FILE PATH", &path_class_ref);

	nwos_generate_new_id(ref);

	nwos_fill_in_common_header(&ptr_path_obj->header.common, ref, &path_class_ref);

	ptr_path_obj->count = length;
	for (i = 0; i < length; i++) ptr_path_obj->storage[i] = path[i];

	nwos_create_reference_list(ref, &ptr_path_obj->header.object.references);

	nwos_crc32_calculate((uint8*) &ptr_path_obj->header.object, sizeof(ObjectHeader), ptr_path_obj->header.common.header_chksum);

	nwos_crc32_calculate((uint8*) &ptr_path_obj->count, sizeof(C_struct_File_Path) + length - sizeof(EveryObject), ptr_path_obj->header.common.data_chksum);

	nwos_write_object_to_disk(ref, kludge, sizeof(C_struct_File_Path) + length);

	nwos_add_to_references(ref, &path_class_ref);

	ptr_path_obj = malloc(sizeof(C_struct_File_Path) + length);
	nwos_read_object_from_disk(ref, ptr_path_obj, sizeof(C_struct_File_Path) + length);
	assert(memcmp(kludge, ptr_path_obj, sizeof(C_struct_File_Path) + length) == 0);

	memset(kludge, 0, sizeof(kludge));  /* clear it */
	nwos_read_variable_sized_object_from_disk(ref, kludge, sizeof(kludge), &get_path_object_size);  /* read the other way */
	assert(memcmp(ptr_path_obj, kludge, sizeof(C_struct_File_Path) + length) == 0);

	free(ptr_path_obj);
	ptr_path_obj = NULL;

	result = CREATED_NEW;
    }

    return result;
}



bool nwos_file_path_to_string(ObjRef* ref, char* string, size_t size)
{
    uint8 kludge[FILE_BLOCK_SIZE];
    C_struct_File_Path* ptr_path_obj;
    int i;
    size_t path_obj_size;

    /* read the path object into the kludge buffer and then after we know what size it is malloc space and copy it there */
    nwos_read_variable_sized_object_from_disk(ref, kludge, sizeof(kludge), &get_path_object_size);

    /* point to the kludge buffer temporarily to get the size of the object */
    ptr_path_obj = (C_struct_File_Path*)kludge;

    path_obj_size = sizeof(C_struct_File_Path) + ptr_path_obj->count;
    /* printf("path_obj_size: %d\n", path_obj_size); */
    assert(path_obj_size > sizeof(C_struct_File_Path));

    /* now we have the size allocate memory for it and copy it there */
    ptr_path_obj = malloc(path_obj_size);
    memcpy(ptr_path_obj, kludge, path_obj_size);

    for (i = 0; i < ptr_path_obj->count; i++) 
    {
	assert(i < size);
	string[i] = ptr_path_obj->storage[i];
    }
    string[i] = '\0';

    free(ptr_path_obj);

    return true;
}


bool nwos_file_path_to_path_and_file_association(ObjRef* path_ref, int index, ObjRef* assoc_ref)
{
    uint8 kludge[FILE_BLOCK_SIZE];
    C_struct_File_Path* ptr_path_obj = (C_struct_File_Path*)kludge;
    ReferenceList* ref_list;
    int num_refs;
    ObjRef object_class;
    ObjRef assoc_class_ref;
    C_struct_Path_And_File_Association assoc_obj;
    int i;
    int count;

    if (!nwos_find_private_class_definition("PATH AND FILE ASSOCIATION", &assoc_class_ref)) return false;


    /* read the path object into the kludge buffer and then after we know what size it is malloc space and copy it there */
    nwos_read_variable_sized_object_from_disk(path_ref, kludge, sizeof(kludge), &get_path_object_size);

    ref_list = nwos_malloc_reference_list(&ptr_path_obj->header.object.references);

    num_refs = ref_list->common_header.num_refs;

    count = 0;

    for (i = 0; i < num_refs; i++)
    {
	nwos_get_object_class(&ref_list->references[i], &object_class);

	if (is_same_object(&object_class, &assoc_class_ref))
	{
	    nwos_read_object_from_disk(&ref_list->references[i], &assoc_obj, sizeof(assoc_obj));

	    if (count == index)    /* this is the one we want */
	    {
		copy_reference(assoc_ref, &ref_list->references[i]);
		break;
	    }
	    else
	    {
		count++;
	    }
	}
    }

    //    memcpy(file_ref, &ptr_path_obj->file, sizeof(*file_ref));

    nwos_free_reference_list(ref_list);
    ref_list = NULL;

    return (i < num_refs);
}


/*-------------------------------------------------------------------------------------------------------------------*/
/* MD5 object stuff */
/*-------------------------------------------------------------------------------------------------------------------*/

bool nwos_find_md5(uint8 md5sum[16], ObjRef* ref)
{
    C_struct_Class_Definition class_def_obj;
    ObjRef md5_class_ref;
    ObjRef object_class;
    ReferenceList* ref_list;
    int num_refs;
    C_struct_MD5sum md5_object;
    int i;

    void_reference(ref);  /* in case we don't find it */


    /* for now return false if the md5sum private class doesn't exist yet */

    if (!nwos_find_private_class_definition("MD5SUM", &md5_class_ref)) return false;

    nwos_read_class_definition(&md5_class_ref, &class_def_obj);

    ref_list = nwos_malloc_reference_list(&class_def_obj.header.object.references);

    num_refs = ref_list->common_header.num_refs;

    /* printf("num_refs: %d\n", num_refs); */

    for (i = 0; i < num_refs; i++)
    {
	nwos_get_object_class(&ref_list->references[i], &object_class);

	if (is_same_object(&object_class, &md5_class_ref))
	{
	    nwos_read_object_from_disk(&ref_list->references[i], &md5_object, sizeof(md5_object));

	    if (memcmp(md5_object.md5sum, md5sum, sizeof(md5sum)) == 0)   /* found a match */
	    {
		memcpy(ref, &ref_list->references[i], sizeof(ObjRef));
		break;
	    }
	}
    }

    nwos_free_reference_list(ref_list);
    ref_list = NULL;

    return (i != num_refs);   /* return true if we found it */
}




/* Find existing MD5sum or create new */

ObjCreateResult nwos_create_md5(uint8 md5sum[16], ObjRef* ref)
{
    ObjRef md5_class_ref;
    ObjCreateResult result = FOUND_EXISTING;
    C_struct_MD5sum md5_object;
    C_struct_MD5sum* ptr_md5_obj;


    /* first find out if we already have this md5 */

    if (!nwos_find_md5(md5sum, ref))   /* didn't find it */
    {
	nwos_find_or_create_private_class_definition("MD5SUM", &md5_class_ref);

	memset(&md5_object, 0, sizeof(md5_object));

	nwos_generate_new_id(ref);

	nwos_fill_in_common_header(&md5_object.header.common, ref, &md5_class_ref);

	memcpy(md5_object.md5sum, md5sum, sizeof(md5_object.md5sum));

	nwos_create_reference_list(ref, &md5_object.header.object.references);

	nwos_crc32_calculate((uint8*) &md5_object.header.object, sizeof(ObjectHeader), md5_object.header.common.header_chksum);

	nwos_crc32_calculate((uint8*) &md5_object.md5sum, sizeof(C_struct_MD5sum) - sizeof(EveryObject), md5_object.header.common.data_chksum);

	nwos_write_object_to_disk(ref, &md5_object, sizeof(md5_object));

	nwos_add_to_references(ref, &md5_class_ref);

	ptr_md5_obj = malloc(sizeof(C_struct_MD5sum));
	nwos_read_object_from_disk(ref, ptr_md5_obj, sizeof(C_struct_MD5sum));
	assert(memcmp(&md5_object, ptr_md5_obj, sizeof(C_struct_MD5sum)) == 0);
	free(ptr_md5_obj);
	ptr_md5_obj = NULL;

	result = CREATED_NEW;
    }

    return result;
}



/*-------------------------------------------------------------------------------------------------------------------*/
/* SHA1 object stuff */
/*-------------------------------------------------------------------------------------------------------------------*/

bool nwos_find_sha1(uint8 sha1sum[20], ObjRef* ref)
{
    C_struct_Class_Definition class_def_obj;
    ObjRef sha1_class_ref;
    ObjRef object_class;
    ReferenceList* ref_list;
    int num_refs;
    C_struct_SHA1sum sha1_object;
    int i;

    void_reference(ref);  /* in case we don't find it */


    /* for now return false if the sha1sum private class doesn't exist yet */

    if (!nwos_find_private_class_definition("SHA1SUM", &sha1_class_ref)) return false;

    nwos_read_class_definition(&sha1_class_ref, &class_def_obj);

    ref_list = nwos_malloc_reference_list(&class_def_obj.header.object.references);

    num_refs = ref_list->common_header.num_refs;

    /* printf("num_refs: %d\n", num_refs); */

    for (i = 0; i < num_refs; i++)
    {
	nwos_get_object_class(&ref_list->references[i], &object_class);

	if (is_same_object(&object_class, &sha1_class_ref))
	{
	    nwos_read_object_from_disk(&ref_list->references[i], &sha1_object, sizeof(sha1_object));

	    if (memcmp(sha1_object.sha1sum, sha1sum, sizeof(sha1sum)) == 0)   /* found a match */
	    {
		memcpy(ref, &ref_list->references[i], sizeof(ObjRef));
		break;
	    }
	}
    }

    nwos_free_reference_list(ref_list);
    ref_list = NULL;

    return (i != num_refs);   /* return true if we found it */
}




/* Find existing SHA1sum or create new */

ObjCreateResult nwos_create_sha1(uint8 sha1sum[20], ObjRef* ref)
{
    ObjRef sha1_class_ref;
    ObjCreateResult result = FOUND_EXISTING;
    C_struct_SHA1sum sha1_object;
    C_struct_SHA1sum* ptr_sha1_obj;


    /* first find out if we already have this sha1 */

    if (!nwos_find_sha1(sha1sum, ref))   /* didn't find it */
    {
	nwos_find_or_create_private_class_definition("SHA1SUM", &sha1_class_ref);

	memset(&sha1_object, 0, sizeof(sha1_object));

	nwos_generate_new_id(ref);

	nwos_fill_in_common_header(&sha1_object.header.common, ref, &sha1_class_ref);

	memcpy(sha1_object.sha1sum, sha1sum, sizeof(sha1_object.sha1sum));

	nwos_create_reference_list(ref, &sha1_object.header.object.references);

	nwos_crc32_calculate((uint8*) &sha1_object.header.object, sizeof(ObjectHeader), sha1_object.header.common.header_chksum);

	nwos_crc32_calculate((uint8*) &sha1_object.sha1sum, sizeof(C_struct_SHA1sum) - sizeof(EveryObject), sha1_object.header.common.data_chksum);

	nwos_write_object_to_disk(ref, &sha1_object, sizeof(sha1_object));

	nwos_add_to_references(ref, &sha1_class_ref);

	ptr_sha1_obj = malloc(sizeof(C_struct_SHA1sum));
	nwos_read_object_from_disk(ref, ptr_sha1_obj, sizeof(C_struct_SHA1sum));
	assert(memcmp(&sha1_object, ptr_sha1_obj, sizeof(C_struct_SHA1sum)) == 0);
	free(ptr_sha1_obj);
	ptr_sha1_obj = NULL;

	result = CREATED_NEW;
    }

    return result;
}



/*-------------------------------------------------------------------------------------------------------------------*/
/* File object stuff */
/*-------------------------------------------------------------------------------------------------------------------*/

static uint32 checksum_file(char* path, uint8 md5_digest[16], uint8 sha1_digest[20])
{
    FILE* fp;
    MD5_CTX md5_context;    /* MD5 checksum context */
    struct sha1_ctx sha1_context;
    size_t bytes_read;
    uint8 buffer[4096];
    uint32 file_length;

    fp = fopen(path, "rb");

    file_length = 0;

    MD5Init(&md5_context);   /* initialize the MD5 checksum context */
    sha1_init_ctx(&sha1_context);

    bytes_read = sizeof(buffer);   /* go through loop at least once */

    while (bytes_read == sizeof(buffer))
    {
	bytes_read = fread(buffer, sizeof(uint8), sizeof(buffer), fp);

	if (bytes_read < sizeof(buffer))    /* eof or error */
	{
	    if (ferror(fp))
	    {
		perror(path);
		exit(1);            /* FIX THIS - needs to clean up and exit gracefully!! */
	    }
	}

	file_length = file_length + bytes_read;

	if (bytes_read > 0)   /* we have data to store */
	{
	    MD5Update(&md5_context, buffer, (unsigned)bytes_read);    /* include this data in the md5 checksum */
	    sha1_process_bytes(buffer, bytes_read, &sha1_context);    /* include this data in the sha1 checksum */
	}
    }

    fclose(fp);

    MD5Final(md5_digest, &md5_context);   /* finish computing the md5 sum */
    sha1_finish_ctx(&sha1_context, sha1_digest);

    return file_length;
}



/* find a matching file from the md5sum */

bool nwos_find_matching_file_from_size_md5_sha1(uint32 file_size, uint8 md5_digest[16], uint8 sha1_digest[20], ObjRef* ref)
{
    ObjRef file_class_ref;
    ObjRef object_class;
    ObjRef md5sum_ref;
    ObjRef sha1sum_ref;
    uint32 size;
    C_struct_SHA1sum sha1_obj;
    C_struct_File file_obj;
    ReferenceList* ref_list;
    int num_refs;
    int i;
    bool result = false;

    void_reference(ref);  /* in case we don't find it */


    if (!nwos_find_private_class_definition("FILE", &file_class_ref)) return false;

    if (nwos_find_md5(md5_digest, &md5sum_ref) && nwos_find_sha1(sha1_digest, &sha1sum_ref))
    {
	if (verbose)
	{
	    printf("Found existing MD5: %02x%02x%02x%02x and SHA1: %02x%02x%02x%02x - ", 
		   md5sum_ref.id[0], md5sum_ref.id[1], md5sum_ref.id[2], md5sum_ref.id[3],
		   sha1sum_ref.id[0], sha1sum_ref.id[1], sha1sum_ref.id[2], sha1sum_ref.id[3]);
	}

	nwos_read_object_from_disk(&sha1sum_ref, &sha1_obj, sizeof(sha1_obj));
	ref_list = nwos_malloc_reference_list(&sha1_obj.header.object.references);
	num_refs = ref_list->common_header.num_refs;

	for (i = 0; i < num_refs; i++)
	{
	    nwos_get_object_class(&ref_list->references[i], &object_class);

	    if (is_same_object(&object_class, &file_class_ref))
	    {
		nwos_read_object_from_disk(&ref_list->references[i], &file_obj, sizeof(file_obj));

		assert(is_same_object(&sha1sum_ref, &file_obj.sha1sum));

		size = ((uint32)file_obj.size[0] << 24) | ((uint32)file_obj.size[1] << 16)
		                      | ((uint32)file_obj.size[2] << 8) | (uint32)file_obj.size[3];

		if (verbose) printf("size: %u\n", size);

		if (size == file_size && is_same_object(&file_obj.md5sum, &md5sum_ref))
		{
		    copy_reference(ref, &ref_list->references[i]);   /* use existing object instead */
		    break;
		}
	    }
	}

	nwos_free_reference_list(ref_list);
	ref_list = NULL;

	result = i < num_refs;
    }

    return result;
}


/* check to see if the file given has the data stored in the system */

bool nwos_file_is_stored(ObjRef* assoc_ref)
{
    C_struct_Path_And_File_Association assoc_obj;
    C_struct_File file_obj;

    nwos_read_object_from_disk(assoc_ref, &assoc_obj, sizeof(assoc_obj));

    nwos_read_object_from_disk(&assoc_obj.file, &file_obj, sizeof(file_obj));

    /* if the block_list is not empty then the data is stored */

    return !is_void_reference(&file_obj.block_list);
}


/* Create a file record (doesn't store data of file, just time, md5, etc.) */

ObjCreateResult nwos_create_file_without_storing_data(char* directory, char* file_name, ObjRef* ref)
{
    size_t length;
    struct stat stat_struct;
    ObjRef file_class_ref;
    ObjRef path_ref;
    ObjRef file_ref;
    uint32 file_length;
    C_struct_File file_obj;
    TimeStamp mod_time;
    uint8 md5_digest[16];
    uint8 sha1_digest[20];
    int i;
    char path[PATH_MAX];

    strncpy(path, directory, PATH_MAX);
    if (path[strlen(path) - 1] != '/')   /* the directory didn't end with a '/' so add one */
    {
	strncat(path, "/", PATH_MAX);
    }
    strncat(path, file_name, PATH_MAX);


    if (stat(path, &stat_struct) != 0)
    {
	perror(path);
	return ERROR_OCCURRED;       /* FIX THIS to handle this more gracefully */
    }

    if (stat_struct.st_size == 0)
    {
	fprintf(stderr, "Cannot handle empty files at this time\n");
	return ERROR_OCCURRED;
    }

    if (stat_struct.st_size > 2147483647)
    {
	fprintf(stderr, "Cannot handle files larger than 2GB at this time\n");
	return ERROR_OCCURRED;
    }


    length = strlen(file_name);

    assert(length <= 255);
    
    printf("creating file record: %s ", file_name);
    fflush(stdout);

    file_length = checksum_file(path, md5_digest, sha1_digest);

    printf("length: %u\n", file_length);

    nwos_convert_timespec_to_time_stamp(&stat_struct.st_mtimespec, mod_time);

    printf("MD5 sum: ");
    for (i = 0; i < 16; i++) printf("%02x", md5_digest[i]);
    printf("\n");

    printf("SHA1 sum: ");
    for (i = 0; i < 20; i++) printf("%02x", sha1_digest[i]);
    printf("\n");

    fflush(stdout);


    /* see if there is a file that matches completely */

    if (!nwos_find_matching_file_from_size_md5_sha1(file_length, md5_digest, sha1_digest, &file_ref))
    {
	nwos_generate_new_id(&file_ref);

	memset(&file_obj, 0, sizeof(file_obj));  /* zero it out */

	nwos_find_or_create_private_class_definition("FILE", &file_class_ref);

	nwos_fill_in_common_header(&file_obj.header.common, &file_ref, &file_class_ref);

	/* store file size as 4 bytes in big endian order */
	file_obj.size[0] = file_length >> 24;
	file_obj.size[1] = file_length >> 16;
	file_obj.size[2] = file_length >> 8;
	file_obj.size[3] = file_length;

	nwos_create_md5(md5_digest, &file_obj.md5sum);

	nwos_create_sha1(sha1_digest, &file_obj.sha1sum);

	nwos_create_reference_list(&file_ref, &file_obj.header.object.references);

	nwos_crc32_calculate((uint8*) &file_obj.header.object, sizeof(ObjectHeader), file_obj.header.common.header_chksum);
	nwos_crc32_calculate((uint8*) &file_obj.size[0], sizeof(C_struct_File) - sizeof(EveryObject), file_obj.header.common.data_chksum);

	nwos_write_object_to_disk(&file_ref, &file_obj, sizeof(file_obj));

	nwos_add_to_references(&file_ref, &file_obj.md5sum);   /* add to this md5sum object's reference list */
	nwos_add_to_references(&file_ref, &file_obj.sha1sum);   /* add to this sha1sum object's reference list */

	nwos_add_to_references(&file_ref, &file_class_ref);
    }

    /* at this point we have a file object, and we need to either find or create a path object */

    assert(nwos_create_file_path(file_name, &path_ref) != ERROR_OCCURRED);

    return nwos_find_or_create_path_and_file_association(&path_ref, &file_ref, mod_time, ref);
}


static uint32 read_file(char* path, uint8 md5_digest[16], uint8 sha1_digest[20], ObjRef* block_list_ref)
{
    FILE* fp;
    size_t bytes_read;
    uint32 file_length;
    ObjRef data_block_ref;
    DataObject data_obj;
    MD5_CTX md5_context;    /* MD5 checksum context */
    struct sha1_ctx sha1_context;
    uint8 ivec[IVEC_SIZE];
    int seq = 0;
    char msg[128];


    memset(ivec, 0, sizeof(ivec));

    /* create a reference list that is a list of the data blocks for the file */
    nwos_create_reference_list(NULL, block_list_ref);

    snprintf(msg, sizeof(msg), "read_file %s  block_list: %02x%02x%02x%02x", path,
	   block_list_ref->id[0], block_list_ref->id[1], block_list_ref->id[2], block_list_ref->id[3]);
    nwos_log(msg);


    /* now open the file and start creating the data block objects */

    fp = fopen(path, "rb");

    file_length = 0;

    MD5Init(&md5_context);   /* initialize the MD5 checksum context */
    sha1_init_ctx(&sha1_context);

    memset(data_obj.flags, 0, sizeof(data_obj.flags));

    bytes_read = DATA_STORAGE_SIZE;   /* go through loop at least once */

    while (bytes_read == DATA_STORAGE_SIZE)
    {
	bytes_read = fread(data_obj.storage, sizeof(uint8), DATA_STORAGE_SIZE, fp);

	if (bytes_read < DATA_STORAGE_SIZE)    /* eof or error */
	{
	    if (ferror(fp))
	    {
		perror(path);
		exit(1);            /* FIX THIS - needs to clean up and exit gracefully!! */
	    }
	}

	file_length = file_length + bytes_read;

	if (bytes_read > 0)   /* we have data to store */
	{
	    MD5Update(&md5_context, data_obj.storage, (unsigned)bytes_read);    /* include this data in the md5 checksum */
	    sha1_process_bytes(data_obj.storage, bytes_read, &sha1_context);    /* include this data in the sha1 checksum */

	    nwos_generate_new_id(&data_block_ref);

	    memcpy(&data_obj.id, &data_block_ref, sizeof(ObjRef));

#if 0
	    printf("data block seq: %d ref: %02x%02x%02x%02x\n", seq, data_block_ref.id[0], data_block_ref.id[1], data_block_ref.id[2], data_block_ref.id[3]);
#endif

	    nwos_crc32_calculate((uint8*) &data_obj.storage, bytes_read, data_obj.checksum);

	    /* fill in the remainder with random data */
	    while (bytes_read < DATA_STORAGE_SIZE) data_obj.storage[bytes_read++] = random();

	    nwos_write_object_to_disk_and_encrypt(&data_block_ref, &data_obj, sizeof(data_obj), ivec, nwos_random_sequence[seq%NUM_STORED_SEQ]);

	    nwos_add_to_reference_list(&data_block_ref, block_list_ref);

	    seq++;
	}
    }

    fclose(fp);

    snprintf(msg, sizeof(msg), "length: %u", file_length);
    nwos_log(msg);
    printf("%s\n", msg);

    MD5Final(md5_digest, &md5_context);   /* finish computing the md5 sum */
    sha1_finish_ctx(&sha1_context, sha1_digest);

    return file_length;
}



/* Create new file - returns FOUND_EXISTING if an exact file match already existed      */
/*          (exact match meaning size, md5, sha1, date, and path are all identical).    */
/*                                                                                      */
/* There are several possible cases of duplication that have to be dealt with:          */
/*   Path is the same but the file is different (file with same name as another file).  */
/*   File is the same but the name is different (copy of a file with a different name). */
/*   Path and File are the same but the date is different.                              */
/*   Everything is identical, but the file isn't stored in the system yet.              */
/*   Everything is identical.                                                           */

ObjCreateResult nwos_create_file(char* path, ObjRef* ref)
{
    size_t length;
    struct stat stat_struct;
    ObjCreateResult result;
    ObjRef file_class_ref;
    ObjRef path_ref;
    ObjRef file_ref;
    uint32 file_length;
    C_struct_File file_obj;
    uint8 md5_digest[16];
    uint8 sha1_digest[20];
    TimeStamp mod_time;

    void_reference(ref);   /* in case it fails */


    if (stat(path, &stat_struct) != 0)
    {
	perror(path);
	return ERROR_OCCURRED;       /* FIX THIS to handle this more gracefully */
    }

    if (stat_struct.st_size == 0)
    {
	fprintf(stderr, "Cannot handle empty files at this time\n");
	return ERROR_OCCURRED;
    }

    if (stat_struct.st_size > 1073741824)
    {
	fprintf(stderr, "Cannot handle files larger than 1GB at this time\n");
	return ERROR_OCCURRED;
    }

    length = strlen(path);

    assert(length <= 255);


    /* first find out if we already have this file stored */

    file_length = checksum_file(path, md5_digest, sha1_digest);

    if (nwos_find_matching_file_from_size_md5_sha1(file_length, md5_digest, sha1_digest, &file_ref))
    {
	return DUPLICATE_FILE;
    }


    /* second find out if we already have this path */

    nwos_push_encryption_level(Encryption_Extreme);     /* in case a new file path is created make new file random */
    
    result = nwos_create_file_path(path, &path_ref);

    nwos_pop_encryption_level();
    
    if (result == FOUND_EXISTING)   /* this file might already exist */
    {
	return DUPLICATE_PATH;
    }


    if (result == CREATED_NEW)   /* it's ok, doesn't exist already */
    {
	printf("creating file: %s ", path);
	fflush(stdout);

	memset(&file_obj, 0, sizeof(file_obj));  /* zero it out */

	nwos_find_or_create_private_class_definition("FILE", &file_class_ref);

	nwos_generate_new_id(&file_ref);

	nwos_fill_in_common_header(&file_obj.header.common, &file_ref, &file_class_ref);

	file_length = read_file(path, md5_digest, sha1_digest, &file_obj.block_list);

	nwos_create_md5(md5_digest, &file_obj.md5sum);
	nwos_create_sha1(sha1_digest, &file_obj.sha1sum);

	nwos_add_to_references(&file_ref, &file_obj.md5sum);   /* add to this md5sum object's reference list */
	nwos_add_to_references(&file_ref, &file_obj.sha1sum);   /* add to this md5sum object's reference list */

	{
	  int i;
	  printf("MD5 sum: ");
	  for (i = 0; i < 16; i++) printf("%02x", md5_digest[i]);
	  printf("\n");
	  printf("SHA1 sum: ");
	  for (i = 0; i < 20; i++) printf("%02x", sha1_digest[i]);
	  printf("\n");
	}

	fflush(stdout);

	/* store file size as 4 bytes in big endian order */
	file_obj.size[0] = file_length >> 24;
	file_obj.size[1] = file_length >> 16;
	file_obj.size[2] = file_length >> 8;
	file_obj.size[3] = file_length;

	nwos_create_reference_list(&file_ref, &file_obj.header.object.references);

	nwos_crc32_calculate((uint8*) &file_obj.header.object, sizeof(ObjectHeader), file_obj.header.common.header_chksum);

	nwos_crc32_calculate((uint8*) &file_obj.size[0], sizeof(C_struct_File) - sizeof(EveryObject), file_obj.header.common.data_chksum);

	nwos_write_object_to_disk(&file_ref, &file_obj, sizeof(file_obj));

	nwos_add_to_references(&file_ref, &file_class_ref);

#if 0
	ptr_file_obj = malloc(sizeof(C_struct_File) + num_blocks * sizeof(ObjRef));
	nwos_read_object_from_disk(ref, ptr_file_obj, sizeof(C_struct_File) + num_blocks * sizeof(ObjRef));
	assert(memcmp(kludge, ptr_file_obj, sizeof(C_struct_File) + num_blocks * sizeof(ObjRef)) == 0);

	memset(kludge, 0, sizeof(kludge));  /* clear it */
	nwos_read_variable_sized_object_from_disk(ref, kludge, sizeof(kludge), &get_file_object_size);  /* read the other way */
	assert(memcmp(ptr_file_obj, kludge, sizeof(C_struct_File_Path) + length) == 0);
#endif

	nwos_convert_timespec_to_time_stamp(&stat_struct.st_mtimespec, mod_time);

	assert(nwos_find_or_create_path_and_file_association(&path_ref, &file_ref, mod_time, ref) == CREATED_NEW);
    }

    return result;
}


/* Check a file's md5sum */

CheckFileResult nwos_check_file_checksums(char* path)
{
    size_t length;
    struct stat stat_struct;
    FILE* fp;
    size_t bytes_read;
    ObjRef path_ref;
    ObjRef file_ref;
    uint8 buffer[4096];
    uint32 file_length;
    int num_blocks;
    C_struct_File file_obj;
    MD5_CTX md5_context;    /* MD5 checksum context */
    struct sha1_ctx sha1_context;
    C_struct_MD5sum md5_object;
    C_struct_SHA1sum sha1_object;
    uint8 md5_digest[16];
    uint8 sha1_digest[20];
    TimeStamp mod_time;
    CheckFileResult result = File_Not_Found;

    if (stat(path, &stat_struct) != 0)
    {
	perror(path);
	return ERROR_OCCURRED;       /* FIX THIS to handle this more gracefully */
    }

    if (stat_struct.st_size == 0)
    {
	fprintf(stderr, "Cannot handle empty files at this time\n");
	return ERROR_OCCURRED;
    }

    if (stat_struct.st_size > 2147483647)
    {
	fprintf(stderr, "Cannot handle files larger than 2GB at this time\n");
	return ERROR_OCCURRED;
    }

    length = strlen(path);

    assert(length <= 255);
    
    /* first find out if we already have this path */

    assert(false);   /* need to fix file_path to assoc below */

    if (nwos_find_file_path(path, &path_ref))   /* it's ok, found the file */
    {
#if 0
	nwos_file_path_to_file(&path_ref, &file_ref);
#endif
	nwos_read_object_from_disk(&file_ref, &file_obj, sizeof(file_obj));  /* read the file object */

	file_length = (file_obj.size[0] << 24) |(file_obj.size[1] << 16) | (file_obj.size[2] << 8) | file_obj.size[3];

	printf("file size: %u\n", file_length);
	/* now open the file and start creating the data block objects */

	fp = fopen(path, "rb");

	file_length = 0;
	num_blocks = 0;

	MD5Init(&md5_context);   /* initialize the MD5 checksum context */
	sha1_init_ctx(&sha1_context);   /* and the SHA1 context */

	bytes_read = sizeof(buffer);   /* go through loop at least once */

	while (bytes_read == sizeof(buffer))
	{
	    bytes_read = fread(buffer, sizeof(uint8), sizeof(buffer), fp);

	    if (bytes_read < sizeof(buffer))    /* eof or error */
	    {
		if (ferror(fp))
		{
		    perror(path);
		    exit(1);            /* FIX THIS - needs to clean up and exit gracefully!! */
		}
	    }

	    file_length = file_length + bytes_read;

	    if (bytes_read > 0)   /* we have data to store */
	    {
		MD5Update(&md5_context, buffer, (unsigned)bytes_read);    /* include this data in the md5 checksum */
		sha1_process_bytes(buffer, bytes_read, &sha1_context);    /* and in the sha1 checksum */
	    }
	}

	fclose(fp);

	printf("length: %u\n", file_length);

	MD5Final(md5_digest, &md5_context);   /* finish computing the md5 sum */
	sha1_finish_ctx(&sha1_context, sha1_digest);  /* and the sha1 sum */


	fflush(stdout);

	if (file_length != ((file_obj.size[0] << 24) | (file_obj.size[1] << 16) | (file_obj.size[2] << 8) | file_obj.size[3]))
	{
	    printf("file size mismatch\n");
	}

	nwos_convert_timespec_to_time_stamp(&stat_struct.st_mtimespec, mod_time);

	assert(false);
#if 0
	if (memcmp(file_obj.modification_time, mod_time, sizeof(TimeStamp)) != 0)
	{
	    printf("time mismatch\n");
	}
#endif
	nwos_read_object_from_disk(&file_obj.md5sum, &md5_object, sizeof(md5_object));

	if (memcmp(md5_digest, md5_object.md5sum, sizeof(md5_digest)) == 0)
	{
	    int j;
	    printf("MD5 sum OK: ");
	    for (j = 0; j < 16; j++) printf("%02x", md5_digest[j]);
	    printf("\n");

	    result = Checksums_Match;    /* found file and MD5 was ok */
	}
	else
	{
	    int j;
	    printf("MD5 checksum error, expected: ");
	    for (j = 0; j < 16; j++) printf("%02x", md5_object.md5sum[j]);
	    printf("\n                    received: ");
	    for (j = 0; j < 16; j++) printf("%02x", md5_digest[j]);
	    printf("\n");

	    result = MD5_Sum_Mismatch;    /* found file but MD5 was bad */
	}


	nwos_read_object_from_disk(&file_obj.sha1sum, &sha1_object, sizeof(sha1_object));

	if (memcmp(sha1_digest, sha1_object.sha1sum, sizeof(sha1_digest)) == 0)
	{
	    int j;
	    printf("SHA1 sum OK: ");
	    for (j = 0; j < 20; j++) printf("%02x", sha1_digest[j]);
	    printf("\n");

	    /* MD5 compare set result, since SHA1 was ok leave it */
	}
	else
	{
	    int j;
	    printf("SHA1 checksum error, expected: ");
	    for (j = 0; j < 20; j++) printf("%02x", sha1_object.sha1sum[j]);
	    printf("\n                    received: ");
	    for (j = 0; j < 20; j++) printf("%02x", sha1_digest[j]);
	    printf("\n");

	    if (result == Checksums_Match)
	    {
		result = SHA1_Sum_Mismatch;    /* found file but SHA1 was bad */
	    }
	    else
	    {
		result = Checksums_Mismatch;
	    }
	}
    }

    return result;   /* didn't find file */
}



/* Read file */

bool nwos_restore_file(ObjRef* assoc_ref, char* path)
{
    bool result = true;
    FILE* fp;
    int i;
    uint32 file_length;
    uint32 num_blocks;
    C_struct_File file_obj;
    C_struct_Path_And_File_Association assoc_obj;
    DataObject data_obj;
    struct timeval tv[2];      /* [0] is access time and [1] is modification time, see man utimes */
    MD5_CTX md5_context;           /* MD5 checksum context */
    struct sha1_ctx sha1_context;
    uint8 md5_digest[16];
    uint8 sha1_digest[20];
    C_struct_MD5sum md5_object;
    C_struct_SHA1sum sha1_object;
    ReferenceList* ref_list_ptr;
    uint8 ivec[IVEC_SIZE];
    uint8 chksum[4];

    memset(ivec, 0, sizeof(ivec));

    nwos_read_object_from_disk(assoc_ref, &assoc_obj, sizeof(assoc_obj));  /* read the path and file assoc. object */

    nwos_read_object_from_disk(&assoc_obj.file, &file_obj, sizeof(file_obj));  /* read the file object */

    file_length = (file_obj.size[0] << 24) |(file_obj.size[1] << 16) | (file_obj.size[2] << 8) | file_obj.size[3];

    printf("file size: %u\n", file_length);

    ref_list_ptr = nwos_malloc_reference_list(&file_obj.block_list);

    num_blocks = (file_length + DATA_STORAGE_SIZE - 1) / DATA_STORAGE_SIZE;

    assert(num_blocks == ref_list_ptr->common_header.num_refs);

//    printf("num_blocks: %u  block_list: %02x%02x%02x%02x\n", num_blocks,
//	   file_obj.block_list.id[0],
//	   file_obj.block_list.id[1],
//	   file_obj.block_list.id[2],
//	   file_obj.block_list.id[3]);


    /* now open the file and start writing the data block objects */

    fp = fopen(path, "wb");

    if (fp == NULL)
    {
	perror(path);
	exit(1);
    }
    else  /* file opened ok */
    {
	i = 0;

	MD5Init(&md5_context);   /* initialize the MD5 checksum context */
	sha1_init_ctx(&sha1_context);   /* and the SHA1 context */

	while (file_length > DATA_STORAGE_SIZE)   /* do all except the last block */
	{
	    nwos_read_object_from_disk_and_decrypt(&ref_list_ptr->references[i], &data_obj, sizeof(data_obj), ivec, nwos_random_sequence[i%NUM_STORED_SEQ]);

	    nwos_crc32_calculate((uint8*) &data_obj.storage, sizeof(data_obj.storage), chksum);

	    if (memcmp(chksum, data_obj.checksum, sizeof(chksum)) != 0)
	    {
		printf("checksum error in file block - sequence: %d\n", i);
		exit(1);
	    }

	    MD5Update(&md5_context, data_obj.storage, DATA_STORAGE_SIZE);    /* include this data in the md5 checksum */
	    sha1_process_bytes(data_obj.storage, DATA_STORAGE_SIZE, &sha1_context);    /* and in the sha1 checksum */

	    if (fwrite(data_obj.storage, sizeof(uint8), DATA_STORAGE_SIZE, fp) != DATA_STORAGE_SIZE)
	    {
		perror(path);
		exit(1);
	    }

	    file_length -= DATA_STORAGE_SIZE;

	    i++;
	}

	nwos_read_object_from_disk_and_decrypt(&ref_list_ptr->references[i], &data_obj, sizeof(data_obj), ivec, nwos_random_sequence[i%NUM_STORED_SEQ]);

	nwos_crc32_calculate((uint8*) &data_obj.storage, file_length, chksum);

	if (memcmp(chksum, data_obj.checksum, sizeof(chksum)) != 0)
	{
	    printf("checksum error in file block - sequence: %d\n", i);
	    exit(1);
	}

	MD5Update(&md5_context, data_obj.storage, file_length);    /* include this data in the md5 checksum */
	sha1_process_bytes(data_obj.storage, file_length, &sha1_context);    /* and in the sha1 checksum */

	if (fwrite(data_obj.storage, sizeof(uint8), file_length, fp) != file_length)
	{
	    perror(path);
	    exit(1);
	}

	if (fclose(fp) != 0)
	{
	    perror(path);
	    exit(1);
	}

	if (result)   /* the file was ok, restore it's modification time */
	{
	    nwos_convert_time_stamp_to_timeval(assoc_obj.modification_time, &tv[1]);
	    tv[0] = tv[1];  /* copy modification time into access time since we don't save the access time */
	    utimes(path, tv);
	}

	MD5Final(md5_digest, &md5_context);   /* finish computing the md5 sum */
	sha1_finish_ctx(&sha1_context, sha1_digest);  /* and the sha1 sum */

	if (result)
	{
	    nwos_read_object_from_disk(&file_obj.md5sum, &md5_object, sizeof(md5_object));

	    if (memcmp(md5_digest, md5_object.md5sum, sizeof(md5_digest)) == 0)
	    {
		int j;
		printf("MD5 sum OK: ");
		for (j = 0; j < 16; j++) printf("%02x", md5_digest[j]);
		printf("\n");
	    }
	    else
	    {
		int j;
		printf("MD5 checksum error, expected: ");
		for (j = 0; j < 16; j++) printf("%02x", md5_object.md5sum[j]);
		printf("\n                    received: ");
		for (j = 0; j < 16; j++) printf("%02x", md5_digest[j]);
		printf("\n");

		result = false;
	    }

	    nwos_read_object_from_disk(&file_obj.sha1sum, &sha1_object, sizeof(sha1_object));

	    if (memcmp(sha1_digest, sha1_object.sha1sum, sizeof(sha1_digest)) == 0)
	    {
		int j;
		printf("SHA1 sum OK: ");
		for (j = 0; j < 20; j++) printf("%02x", sha1_digest[j]);
		printf("\n");
	    }
	    else
	    {
		int j;
		printf("SHA1 checksum error, expected: ");
		for (j = 0; j < 20; j++) printf("%02x", sha1_object.sha1sum[j]);
		printf("\n                    received: ");
		for (j = 0; j < 20; j++) printf("%02x", sha1_digest[j]);
		printf("\n");

		result = false;
	    }
	}

	nwos_free_reference_list(ref_list_ptr);
	ref_list_ptr = NULL;
    }

    return result;
}


/* Compare file - does not verify checksums (just compares the data) */

bool nwos_file_is_identical(ObjRef* assoc_ref, char* path)
{
    bool result = true;
    FILE* fp;
    int i;
    struct stat stat_struct;
    uint32 file_length;
    uint32 num_blocks;
    C_struct_File file_obj;
    C_struct_Path_And_File_Association assoc_obj;
    DataObject data_obj;
    struct timeval tv;
    ReferenceList* ref_list_ptr;
    uint8 ivec[IVEC_SIZE];
    uint8 chksum[4];
    uint8 buffer[DATA_STORAGE_SIZE];

    memset(ivec, 0, sizeof(ivec));

    if (stat(path, &stat_struct) != 0)
    {
	perror(path);
	return false;
    }

    nwos_read_object_from_disk(assoc_ref, &assoc_obj, sizeof(assoc_obj));  /* read the path and file assoc. object */

    nwos_read_object_from_disk(&assoc_obj.file, &file_obj, sizeof(file_obj));  /* read the file object */

    file_length = (file_obj.size[0] << 24) |(file_obj.size[1] << 16) | (file_obj.size[2] << 8) | file_obj.size[3];

    printf("file size: %u\n", file_length);

    if (stat_struct.st_size != file_length)
    {
	fprintf(stderr, "File size mismatch\n");
	return false;
    }

    ref_list_ptr = nwos_malloc_reference_list(&file_obj.block_list);

    num_blocks = (file_length + DATA_STORAGE_SIZE - 1) / DATA_STORAGE_SIZE;

    assert(num_blocks == ref_list_ptr->common_header.num_refs);

//    printf("num_blocks: %u  block_list: %02x%02x%02x%02x\n", num_blocks,
//	   file_obj.block_list.id[0],
//	   file_obj.block_list.id[1],
//	   file_obj.block_list.id[2],
//	   file_obj.block_list.id[3]);


    /* now open the file and start writing the data block objects */

    fp = fopen(path, "rb");

    if (fp == NULL)
    {
	perror(path);
	exit(1);
    }
    else  /* file opened ok */
    {
	i = 0;

	while (file_length > DATA_STORAGE_SIZE)   /* do all except the last block */
	{
	    nwos_read_object_from_disk_and_decrypt(&ref_list_ptr->references[i], &data_obj, sizeof(data_obj), ivec, nwos_random_sequence[i%NUM_STORED_SEQ]);

	    nwos_crc32_calculate((uint8*) &data_obj.storage, sizeof(data_obj.storage), chksum);

	    if (memcmp(chksum, data_obj.checksum, sizeof(chksum)) != 0)
	    {
		printf("checksum error in file block - sequence: %d\n", i);
		exit(1);
	    }

	    if (fread(buffer, sizeof(uint8), sizeof(buffer), fp) != DATA_STORAGE_SIZE)
	    {
		perror(path);
		exit(1);
	    }

	    if (memcmp(&data_obj.storage, buffer, DATA_STORAGE_SIZE) != 0)
	    {
		result = false;
		break;
	    }

	    file_length -= DATA_STORAGE_SIZE;

	    i++;
	}


	if (result)   /* file compared ok thus far */
	{
	    nwos_read_object_from_disk_and_decrypt(&ref_list_ptr->references[i], &data_obj, sizeof(data_obj), ivec, nwos_random_sequence[i%NUM_STORED_SEQ]);

	    nwos_crc32_calculate((uint8*) &data_obj.storage, file_length, chksum);

	    if (memcmp(chksum, data_obj.checksum, sizeof(chksum)) != 0)
	    {
		printf("checksum error in file block - sequence: %d\n", i);
		exit(1);
	    }

	    if (fread(buffer, sizeof(uint8), file_length, fp) != file_length)
	    {
		perror(path);
		exit(1);
	    }

	    if (memcmp(&data_obj.storage, buffer, file_length) != 0)
	    {
		result = false;
	    }

	    if (fclose(fp) != 0)
	    {
		perror(path);
		exit(1);
	    }
	}

	if (result)   /* the file was ok, compare it's modification time */
	{
	    nwos_convert_time_stamp_to_timeval(assoc_obj.modification_time, &tv);

	    if (stat_struct.st_mtimespec.tv_sec != tv.tv_sec || stat_struct.st_mtimespec.tv_nsec / 1000 != tv.tv_usec)
	    {
		printf("Time values mismatch - file %ld.%ld  stored: %ld.%ld\n",
		       stat_struct.st_mtimespec.tv_sec, stat_struct.st_mtimespec.tv_nsec / 1000, tv.tv_sec, tv.tv_usec);
		result = false;
	    }
	}
    }

    nwos_free_reference_list(ref_list_ptr);
    ref_list_ptr = NULL;

    return result;
}


/* Returns true if the file is stored on some backup media */

bool nwos_file_has_backup(ObjRef* assoc_ref)
{
    ObjRef disc_list_class_ref;
    ObjRef disc_copy_class_ref;
    ObjRef class_ref;
    C_struct_Path_And_File_Association assoc_obj;
    C_struct_Disc_Copy copy_obj;
    uint8 kludge[MAX_SIZE_DISC_LIST];
    C_struct_Disc_List* ptr_disc_obj = (C_struct_Disc_List*)kludge;
    ReferenceList* assoc_ref_list_ptr;
    ReferenceList* disc_ref_list_ptr;
    int num_assoc_refs;
    int num_disc_refs;
    int i;
    int j;
    int count;

    if (!nwos_find_private_class_definition("DISC LIST", &disc_list_class_ref)) return false;
    if (!nwos_find_private_class_definition("DISC COPY", &disc_copy_class_ref)) return false;

    nwos_read_object_from_disk(assoc_ref, &assoc_obj, sizeof(assoc_obj));  /* read the path and file assoc. object */

    assoc_ref_list_ptr = nwos_malloc_reference_list(&assoc_obj.header.object.references);

    num_assoc_refs = assoc_ref_list_ptr->common_header.num_refs;

    for (i = 0; i < num_assoc_refs; i++)
    {
	nwos_get_object_class(&assoc_ref_list_ptr->references[i], &class_ref);

	if (is_same_object(&class_ref, &disc_list_class_ref))
	{
	    nwos_read_variable_sized_object_from_disk(&assoc_ref_list_ptr->references[i], kludge, sizeof(kludge), &nwos_get_disc_list_object_size);

	    disc_ref_list_ptr = nwos_malloc_reference_list(&ptr_disc_obj->header.object.references);

	    /* as long as we had to read the object, verify that this file is in the list */

	    count = nwos_decode_variable_sized_count(ptr_disc_obj->count);
	    for (j = 0; j < count; j++)
	    {
		if (is_same_object(assoc_ref, &ptr_disc_obj->files[j])) break;
	    }
	    assert(j < count);   /* something very wrong if it wasn't in the list */

	    /* now find any disc copies that exist */

	    num_disc_refs = disc_ref_list_ptr->common_header.num_refs;

	    for (j = 0; j < num_disc_refs; j++)
	    {
		nwos_get_object_class(&disc_ref_list_ptr->references[j], &disc_copy_class_ref);

		if (is_same_object(&class_ref, &disc_list_class_ref))
		{
		    nwos_read_object_from_disk(&disc_ref_list_ptr->references[j], &copy_obj, sizeof(copy_obj));
		    assert(is_same_object(&assoc_ref_list_ptr->references[i], &copy_obj.disc_list));
		    break;
		}
	    }
	    
	    nwos_free_reference_list(disc_ref_list_ptr);
	    disc_ref_list_ptr = NULL;

	    if (j < num_disc_refs) break;  /* found a disc_copy in references */
	}
    }

    nwos_free_reference_list(assoc_ref_list_ptr);
    assoc_ref_list_ptr = NULL;

    return i < num_assoc_refs;;
}


/*-------------------------------------------------------------------------------------------------------------------*/
/* File association (Path <-> File) object stuff */
/*-------------------------------------------------------------------------------------------------------------------*/

bool nwos_find_path_and_file_association_by_index(ObjRef* path_ref, ObjRef* file_ref, int index, ObjRef* ref)
{
    C_struct_Class_Definition class_def_obj;
    C_struct_Path_And_File_Association assoc_obj;
    ObjRef assoc_class_ref;
    ObjRef object_class;
    ReferenceList* ref_list;
    int num_refs;
    int i;
    int count;

    void_reference(ref);  /* in case we don't find it */


    /* for now just return false if there is no private class definition yet */

    if (!nwos_find_private_class_definition("PATH AND FILE ASSOCIATION", &assoc_class_ref)) return false;


    nwos_read_class_definition(&assoc_class_ref, &class_def_obj);

    ref_list = nwos_malloc_reference_list(&class_def_obj.header.object.references);

    num_refs = ref_list->common_header.num_refs;

    /* printf("num_refs: %d\n", num_refs); */

    count = 0;

    for (i = 0; i < num_refs; i++)
    {
	nwos_get_object_class(&ref_list->references[i], &object_class);

	if (is_same_object(&object_class, &assoc_class_ref))
	{
	    nwos_read_object_from_disk(&ref_list->references[i], &assoc_obj, sizeof(assoc_obj));

	    if (is_same_object(&assoc_obj.file, file_ref) && is_same_object(&assoc_obj.path, path_ref))   /* found a match */
	    {
		if (count == index)
		{
		    memcpy(ref, &ref_list->references[i], sizeof(ObjRef));
		    break;
		}
		else
		{
		    count++;
		}
	    }
	}
    }

    nwos_free_reference_list(ref_list);
    ref_list = NULL;

    return (i != num_refs);   /* return true if we found it */
}


bool nwos_find_path_and_file_association(ObjRef* path_ref, ObjRef* file_ref, TimeStamp mod_time, ObjRef* ref)
{
    C_struct_Class_Definition class_def_obj;
    C_struct_Path_And_File_Association assoc_obj;
    ObjRef assoc_class_ref;
    ObjRef object_class;
    ReferenceList* ref_list;
    int num_refs;
    int i;

    void_reference(ref);  /* in case we don't find it */


    /* for now just return false if there is no private class definition yet */

    if (!nwos_find_private_class_definition("PATH AND FILE ASSOCIATION", &assoc_class_ref)) return false;


    nwos_read_class_definition(&assoc_class_ref, &class_def_obj);

    ref_list = nwos_malloc_reference_list(&class_def_obj.header.object.references);

    num_refs = ref_list->common_header.num_refs;

    /* printf("num_refs: %d\n", num_refs); */

    for (i = 0; i < num_refs; i++)
    {
	nwos_get_object_class(&ref_list->references[i], &object_class);

	if (is_same_object(&object_class, &assoc_class_ref))
	{
	    nwos_read_object_from_disk(&ref_list->references[i], &assoc_obj, sizeof(assoc_obj));

	    if (is_same_object(&assoc_obj.file, file_ref) && is_same_object(&assoc_obj.path, path_ref))   /* found a match */
	    {
		if (mod_time == NULL || memcmp(assoc_obj.modification_time, mod_time, sizeof(mod_time)) == 0)
		{
		    memcpy(ref, &ref_list->references[i], sizeof(ObjRef));
		    break;
		}
	    }
	}
    }

    nwos_free_reference_list(ref_list);
    ref_list = NULL;

    return (i != num_refs);   /* return true if we found it */
}


bool nwos_find_matching_path_and_file_association(char* dir, char* file_name, ObjRef* ref, MatchCode match)
{
    size_t length;
    struct stat stat_struct;
    uint32 file_length;
    uint8 md5_digest[16];
    uint8 sha1_digest[20];
    ObjRef path_ref;
    ObjRef file_ref;
    TimeStamp mod_time;
    bool path_found;
    bool file_found;
    char path[PATH_MAX];

    void_reference(ref);  /* in case we don't find it */

    if (dir == NULL)      /* no directory specified */
    {
	strncpy(path, file_name, PATH_MAX);
    }
    else
    {
	strncpy(path, dir, PATH_MAX);
	if (path[strlen(path) - 1] != '/')   /* the directory didn't end with a '/' so add one */
	{
	    strncat(path, "/", PATH_MAX);
	}
	strncat(path, file_name, PATH_MAX);
    }

    if (stat(path, &stat_struct) != 0)
    {
	perror(path);
	return ERROR_OCCURRED;       /* FIX THIS to handle this more gracefully */
    }

    if (stat_struct.st_size == 0)
    {
	fprintf(stderr, "Cannot handle empty files at this time\n");
	return ERROR_OCCURRED;
    }

    if (stat_struct.st_size > 2147483647)
    {
	fprintf(stderr, "Cannot handle files larger than 2GB at this time\n");
	return ERROR_OCCURRED;
    }


    length = strlen(file_name);

    assert(length <= 255);
    
    file_length = checksum_file(path, md5_digest, sha1_digest);

    path_found = nwos_find_file_path(file_name, &path_ref);

    file_found = nwos_find_matching_file_from_size_md5_sha1(file_length, md5_digest, sha1_digest, &file_ref);

    printf("(found path: %s  file: %s) ", path_found ? "yes" : "no", file_found ? "yes" : "no");

    if (!path_found || !file_found) return false;

    if (match == IgnoreTime)
    {
	return nwos_find_path_and_file_association(&path_ref, &file_ref, NULL, ref);
    }

    nwos_convert_timespec_to_time_stamp(&stat_struct.st_mtimespec, mod_time);

    return nwos_find_path_and_file_association(&path_ref, &file_ref, mod_time, ref);
}



/* Find existing file association or create new */

ObjCreateResult nwos_find_or_create_path_and_file_association(ObjRef* path_ref, ObjRef* file_ref, TimeStamp mod_time, ObjRef* ref)
{
    C_struct_Path_And_File_Association assoc_obj;
    C_struct_Path_And_File_Association temp;
    ObjRef assoc_class_ref;
    ObjCreateResult result = FOUND_EXISTING;


    /* first find out if we already have this association */

    if (!nwos_find_path_and_file_association(path_ref, file_ref, mod_time, ref))   /* didn't find existing association */
    {
	memset(&assoc_obj, 0, sizeof(assoc_obj));  /* zero it out */

	nwos_find_or_create_private_class_definition("PATH AND FILE ASSOCIATION", &assoc_class_ref);

	nwos_generate_new_id(ref);

	nwos_fill_in_common_header(&assoc_obj.header.common, ref, &assoc_class_ref);

	copy_reference(&assoc_obj.path, path_ref);
	copy_reference(&assoc_obj.file, file_ref);
	memcpy(assoc_obj.modification_time, mod_time, sizeof(assoc_obj.modification_time));

	nwos_create_reference_list(ref, &assoc_obj.header.object.references);

	nwos_crc32_calculate((uint8*) &assoc_obj.header.object, sizeof(ObjectHeader), assoc_obj.header.common.header_chksum);

	nwos_crc32_calculate((uint8*) &assoc_obj.path, sizeof(C_struct_Path_And_File_Association) - sizeof(EveryObject), assoc_obj.header.common.data_chksum);

	nwos_write_object_to_disk(ref, &assoc_obj, sizeof(assoc_obj));

	nwos_add_to_references(ref, &assoc_class_ref);
	nwos_add_to_references(ref, file_ref);
	nwos_add_to_references(ref, path_ref);

	/* read back to verify */
	nwos_read_object_from_disk(ref, &temp, sizeof(temp));
	assert(memcmp(&temp, &assoc_obj, sizeof(C_struct_Path_And_File_Association)) == 0);

	result = CREATED_NEW;
    }

    return result;
}


/*-------------------------------------------------------------------------------------------------------------------*/
/* Disc List object stuff */
/*-------------------------------------------------------------------------------------------------------------------*/

size_t nwos_get_disc_list_object_size(void* disc_list_obj)
{
    uint32 count = nwos_decode_variable_sized_count(((C_struct_Disc_List*)disc_list_obj)->count);

    assert(0 < count && count <= MAX_FILES_PER_DISC_LIST);

    return sizeof(C_struct_Disc_List) + count * sizeof(ObjRef);
}


bool nwos_find_disc_list(char id[12], ObjRef* ref)
{
    C_struct_Class_Definition class_def_obj;
    uint8 kludge[MAX_SIZE_DISC_LIST];
    C_struct_Disc_List* ptr_disc_obj = (C_struct_Disc_List*)kludge;
    ObjRef disc_class_ref;
    ObjRef object_class;
    ReferenceList* ref_list;
    int num_refs;
    int i;

    void_reference(ref);  /* in case we don't find it */


    /* for now return false if the private class doesn't exist */

    if (!nwos_find_private_class_definition("DISC LIST", &disc_class_ref)) return false;


    nwos_read_class_definition(&disc_class_ref, &class_def_obj);

    ref_list = nwos_malloc_reference_list(&class_def_obj.header.object.references);

    num_refs = ref_list->common_header.num_refs;

    /* printf("num_refs: %d\n", num_refs); */

    for (i = 0; i < num_refs; i++)
    {
	nwos_get_object_class(&ref_list->references[i], &object_class);

	if (is_same_object(&object_class, &disc_class_ref))
	{
	    nwos_read_variable_sized_object_from_disk(&ref_list->references[i], kludge, sizeof(kludge), &nwos_get_disc_list_object_size);

	    /* remember ptr_disc_obj points to the kludge buffer */

	    if (memcmp(ptr_disc_obj->id, id, sizeof(ptr_disc_obj->id)) == 0)   /* found a match */
	    {
		memcpy(ref, &ref_list->references[i], sizeof(ObjRef));
		break;
	    }
	}
    }

    nwos_free_reference_list(ref_list);
    ref_list = NULL;

    return (i != num_refs);   /* return true if we found it */
}



bool nwos_find_matching_disc_list(ObjRef files[MAX_FILES_PER_DISC_LIST], ObjRef* ref)
{
    C_struct_Class_Definition class_def_obj;
    uint8 kludge[MAX_SIZE_DISC_LIST];
    C_struct_Disc_List* ptr_disc_obj = (C_struct_Disc_List*)kludge;
    ObjRef disc_class_ref;
    ObjRef object_class;
    ReferenceList* ref_list;
    int num_refs;
    int i;
    int j;
    int k;
    int count = 0;

    void_reference(ref);  /* in case we don't find it */


    /* first count the files */

    for (i = 0; i < MAX_FILES_PER_DISC_LIST; i++)
    {
	if (!is_void_reference(&files[i])) count++;
    }

    assert(count > 0);

    /* for now return false if the private disc list class doesn't exist */

    if (!nwos_find_private_class_definition("DISC LIST", &disc_class_ref)) return false;


    nwos_read_class_definition(&disc_class_ref, &class_def_obj);

    ref_list = nwos_malloc_reference_list(&class_def_obj.header.object.references);

    num_refs = ref_list->common_header.num_refs;

    /* printf("num_refs: %d\n", num_refs); */

    for (i = 0; i < num_refs; i++)
    {
	nwos_get_object_class(&ref_list->references[i], &object_class);

	if (is_same_object(&object_class, &disc_class_ref))
	{
	    nwos_read_variable_sized_object_from_disk(&ref_list->references[i], kludge, sizeof(kludge), &nwos_get_disc_list_object_size);

	    /* remember ptr_disc_obj points to the kludge buffer */

	    if (nwos_decode_variable_sized_count(ptr_disc_obj->count) == count)   /* found a possible match */
	    {
		for (j = 0; j < MAX_FILES_PER_DISC_LIST; j++)
		{
		    if (!is_void_reference(&files[j]))
		    {
			for (k = 0; k < count; k++)
			{
			    if (is_same_object(&files[j], &ptr_disc_obj->files[k])) break;
			}

			if (k == count) break;   /* didn't find a match this can't be the list */
		    }
		}

		if (j == MAX_FILES_PER_DISC_LIST)   /* found a match for each one */
		{
		    memcpy(ref, &ref_list->references[i], sizeof(ObjRef));
		    break;
		}
	    }
	}
    }

    nwos_free_reference_list(ref_list);
    ref_list = NULL;

    return (i != num_refs);   /* return true if we found it */
}



/* Find existing disc list or create new */

ObjCreateResult nwos_create_disc_list(char id[12], ObjRef files[MAX_FILES_PER_DISC_LIST], ObjRef* ref)
{
    uint8 kludge[MAX_SIZE_DISC_LIST];
    C_struct_Disc_List* ptr_disc_obj = (C_struct_Disc_List*)kludge;
    ObjRef disc_class_ref;
    ObjCreateResult result = FOUND_EXISTING;
    int i;
    int count = 0;


    /* first find out if we already have this disc */

    if (!nwos_find_disc_list(id, ref))   /* didn't find it */
    {
	memset(kludge, 0, sizeof(kludge));  /* zero it out */

	/* remember ptr_path_obj points to the kludge buffer */

	nwos_find_or_create_private_class_definition("DISC LIST", &disc_class_ref);

	nwos_generate_new_id(ref);

	nwos_fill_in_common_header(&ptr_disc_obj->header.common, ref, &disc_class_ref);

	memcpy(&ptr_disc_obj->id, id, sizeof(ptr_disc_obj->id));

	for (i = 0; i < MAX_FILES_PER_DISC_LIST; i++)
	{
	    if (!is_void_reference(&files[i]))
	    {
		copy_reference(&ptr_disc_obj->files[count], &files[i]);
		count++;

		nwos_add_to_references(ref, &files[i]);
	    }
	}

	nwos_encode_variable_sized_count(count, ptr_disc_obj->count);

	assert(nwos_get_disc_list_object_size(ptr_disc_obj) == sizeof(C_struct_Disc_List) + count * sizeof(ObjRef));

	nwos_create_reference_list(ref, &ptr_disc_obj->header.object.references);

	nwos_crc32_calculate((uint8*) &ptr_disc_obj->header.object, sizeof(ObjectHeader), ptr_disc_obj->header.common.header_chksum);

	nwos_crc32_calculate((uint8*) &ptr_disc_obj->id, nwos_get_disc_list_object_size(ptr_disc_obj) - sizeof(EveryObject), ptr_disc_obj->header.common.data_chksum);

	nwos_write_object_to_disk(ref, kludge, nwos_get_disc_list_object_size(ptr_disc_obj));

	nwos_add_to_references(ref, &disc_class_ref);

	ptr_disc_obj = malloc(sizeof(C_struct_Disc_List) + count * sizeof(ObjRef));
	nwos_read_object_from_disk(ref, ptr_disc_obj, sizeof(C_struct_Disc_List) + count * sizeof(ObjRef));
	assert(memcmp(kludge, ptr_disc_obj, sizeof(C_struct_Disc_List) + count * sizeof(ObjRef)) == 0);

	memset(kludge, 0, sizeof(kludge));  /* clear it */
	nwos_read_variable_sized_object_from_disk(ref, kludge, sizeof(kludge), &nwos_get_disc_list_object_size);  /* read the other way */
	assert(memcmp(ptr_disc_obj, kludge, sizeof(C_struct_Disc_List) + count * sizeof(ObjRef)) == 0);

	free(ptr_disc_obj);
	ptr_disc_obj = NULL;

	result = CREATED_NEW;
    }

    return result;
}


/*-------------------------------------------------------------------------------------------------------------------*/
/* Disc Copy object stuff */
/*-------------------------------------------------------------------------------------------------------------------*/


/* find existing disc copy */

bool nwos_find_disc_copy(ObjRef* disc_list, int copy_num, ObjRef* ref)
{
    C_struct_Class_Definition class_def_obj;
    C_struct_Disc_Copy disc_obj;
    ObjRef disc_class_ref;
    ObjRef object_class;
    ReferenceList* ref_list;
    int num_refs;
    int i;

    void_reference(ref);  /* in case we don't find it */


    /* for now return false if there isn't a disc copy private class definition */

    if (!nwos_find_private_class_definition("DISC COPY", &disc_class_ref)) return false;

    nwos_read_class_definition(&disc_class_ref, &class_def_obj);

    ref_list = nwos_malloc_reference_list(&class_def_obj.header.object.references);

    num_refs = ref_list->common_header.num_refs;

    /* printf("num_refs: %d\n", num_refs); */

    for (i = 0; i < num_refs; i++)
    {
	nwos_get_object_class(&ref_list->references[i], &object_class);

	if (is_same_object(&object_class, &disc_class_ref))
	{
	    nwos_read_object_from_disk(&ref_list->references[i], &disc_obj, sizeof(disc_obj));

	    if (is_same_object(&disc_obj.disc_list, disc_list) && disc_obj.copy_number == copy_num)   /* found a match */
	    {
		memcpy(ref, &ref_list->references[i], sizeof(ObjRef));
		break;
	    }
	}
    }

    nwos_free_reference_list(ref_list);
    ref_list = NULL;

    return (i != num_refs);   /* return true if we found it */
}



/* Find existing disc copy or create new */

ObjCreateResult nwos_create_disc_copy(ObjRef* disc_list, int copy_num, ObjRef* location, ObjRef* ref)
{
    C_struct_Disc_Copy disc_copy_obj;
    C_struct_Disc_Copy* ptr_disc_obj;
    ObjRef disc_class_ref;
    ObjCreateResult result = FOUND_EXISTING;


    /* first find out if we already have this disc copy */

    if (!nwos_find_disc_copy(disc_list, copy_num, ref))   /* didn't find it */
    {
	nwos_find_or_create_private_class_definition("DISC COPY", &disc_class_ref);

	memset(&disc_copy_obj, 0, sizeof(disc_copy_obj));

	nwos_generate_new_id(ref);

	nwos_fill_in_common_header(&disc_copy_obj.header.common, ref, &disc_class_ref);

	copy_reference(&disc_copy_obj.disc_list, disc_list);
	copy_reference(&disc_copy_obj.location, location);
	disc_copy_obj.copy_number = copy_num;

	nwos_create_reference_list(ref, &disc_copy_obj.header.object.references);

	nwos_crc32_calculate((uint8*) &disc_copy_obj.header.object, sizeof(ObjectHeader), disc_copy_obj.header.common.header_chksum);

	nwos_crc32_calculate((uint8*) &disc_copy_obj.disc_list, sizeof(disc_copy_obj) - sizeof(EveryObject), disc_copy_obj.header.common.data_chksum);

	nwos_write_object_to_disk(ref, &disc_copy_obj, sizeof(disc_copy_obj));

	nwos_add_to_references(ref, &disc_class_ref);

	nwos_add_to_references(ref, disc_list);
	nwos_add_to_references(ref, location);

	ptr_disc_obj = malloc(sizeof(C_struct_Disc_Copy));
	nwos_read_object_from_disk(ref, ptr_disc_obj, sizeof(C_struct_Disc_Copy));
	assert(memcmp(&disc_copy_obj, ptr_disc_obj, sizeof(C_struct_Disc_Copy)) == 0);

	free(ptr_disc_obj);
	ptr_disc_obj = NULL;

	result = CREATED_NEW;
    }

    return result;
}



/*-------------------------------------------------------------------------------------------------------------------*/
/* Storage Location object stuff */
/*-------------------------------------------------------------------------------------------------------------------*/


/* find existing storage location */

bool nwos_find_storage_location(char* location, ObjRef* ref)
{
    C_struct_Class_Definition class_def_obj;
    C_struct_Storage_Location location_obj;
    ObjRef location_class_ref;
    ObjRef object_class;
    ObjRef name_ref;
    ReferenceList* ref_list;
    int num_refs;
    int i;

    void_reference(ref);  /* in case we don't find it */


    if (!nwos_find_private_name(location, &name_ref))
    {
	return false;
    }


    /* for now return false if storage location private class definition doesn't exist */

    if (!nwos_find_private_class_definition("STORAGE LOCATION", &location_class_ref)) return false;


    nwos_read_class_definition(&location_class_ref, &class_def_obj);

    ref_list = nwos_malloc_reference_list(&class_def_obj.header.object.references);

    num_refs = ref_list->common_header.num_refs;

    /* printf("num_refs: %d\n", num_refs); */

    for (i = 0; i < num_refs; i++)
    {
	nwos_get_object_class(&ref_list->references[i], &object_class);

	if (is_same_object(&object_class, &location_class_ref))
	{
	    nwos_read_object_from_disk(&ref_list->references[i], &location_obj, sizeof(location_obj));

	    if (is_same_object(&location_obj.name, &name_ref))   /* found a match */
	    {
		memcpy(ref, &ref_list->references[i], sizeof(ObjRef));
		break;
	    }
	}
    }

    nwos_free_reference_list(ref_list);
    ref_list = NULL;

    return (i != num_refs);   /* return true if we found it */
}



/* Find existing disc copy or create new */

ObjCreateResult nwos_create_storage_location(char* location, ObjRef* ref)
{
    C_struct_Storage_Location location_obj;
    C_struct_Storage_Location* ptr_loc_obj;
    ObjRef location_class_ref;
    ObjRef name_ref;
    ObjCreateResult result = FOUND_EXISTING;


    /* first find out if we already have this location */

    if (!nwos_find_storage_location(location, ref))   /* didn't find it */
    {
	nwos_create_name(location, &name_ref);

	nwos_find_or_create_private_class_definition("STORAGE LOCATION", &location_class_ref);

	memset(&location_obj, 0, sizeof(location_obj));

	nwos_generate_new_id(ref);

	nwos_fill_in_common_header(&location_obj.header.common, ref, &location_class_ref);

	copy_reference(&location_obj.name, &name_ref);

	nwos_create_reference_list(ref, &location_obj.header.object.references);

	nwos_crc32_calculate((uint8*) &location_obj.header.object, sizeof(ObjectHeader), location_obj.header.common.header_chksum);

	nwos_crc32_calculate((uint8*) &location_obj.name, sizeof(location_obj) - sizeof(EveryObject), location_obj.header.common.data_chksum);

	nwos_write_object_to_disk(ref, &location_obj, sizeof(location_obj));

	nwos_add_to_references(ref, &location_class_ref);

	nwos_add_to_references(ref, &name_ref);

	ptr_loc_obj = malloc(sizeof(C_struct_Storage_Location));
	nwos_read_object_from_disk(ref, ptr_loc_obj, sizeof(C_struct_Storage_Location));
	assert(memcmp(&location_obj, ptr_loc_obj, sizeof(C_struct_Storage_Location)) == 0);

	free(ptr_loc_obj);
	ptr_loc_obj = NULL;

	result = CREATED_NEW;
    }

    return result;
}



